//------------------------------------------------------------------------
// Math Library for right-handed 3D graphics APIs.
// Copyright (C) 2001-2002 David Poon
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General 
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//------------------------------------------------------------------------

#ifndef MATH_LIBRARY_RIGHT_HANDED_HPP
#define MATH_LIBRARY_RIGHT_HANDED_HPP

#include <cmath>
#include <limits>

extern const float PI;

//------------------------------------------------------------------------
// Math library inlined functions
//------------------------------------------------------------------------

inline bool CloseEnough(const float &f1, const float &f2)
{
	// Determines whether the two floating point values f1 and f2 are
	// close enough together that they can be considered equal.
	// This function is taken from Pete Becker's "The Journeyman's Shop"
	// column published in C/C++ Users Journal, December 2000. It is 
	// presented here slightly modified to deal exclusively with floats.
	return fabs((f1 - f2) / ((f2 == 0.f) ? 1.f : f2)) 
		< std::numeric_limits<float>::epsilon();
}

inline float Degrees2Radians(float angle)
{
	// Converts the angle (in degrees) to radians.
	return (angle * PI) / 180.f;
}

inline float Radians2Degrees(float angle)
{
	// Converts the angle (in radians) to degrees.
	return (angle * 180.f) / PI;
}

//------------------------------------------------------------------------
// Definitions of common 3D data structures for use in 3D graphics.
// These data structures are suitable for graphics APIs that use a
// right-handed cartesian coordinate system (i.e., OpenGL).
//------------------------------------------------------------------------

class Matrix;
struct Point;
struct Vector;
struct Quaternion;
struct Plane;

class Matrix
{
public:
	Matrix();
	Matrix(const Matrix &rhs);
	~Matrix();

	float &operator()(int row, int col) {return elementAt(row, col);}
	const float &operator()(int row, int col) const {return elementAt(row, col);}

	void add(const Matrix &m);
	void add(const Matrix &m1, const Matrix &m2);
	void subtract(const Matrix &m);
	void subtract(const Matrix &m1, const Matrix &m2);
	void multiply(float scalar);
	void multiply(const Matrix &m);
	void multiply(const Matrix &m1, const Matrix &m2);
		
	Matrix &operator=(const Matrix &rhs);
	bool operator==(const Matrix &rhs) const;
	bool operator!=(const Matrix &rhs) const;
	
	operator float*() {return m_mtx;}
	operator const float*() const {return m_mtx;}

	bool inverse(const Matrix &m);
	void loadIdentity();
	void transpose(const Matrix &m);
	
	void rotate(float angle, Vector &axis);
	void rotate(const Vector &from, const Vector &to);
	void scale(float sx, float sy, float sz);
	void translate(float tx, float ty, float tz);

private:
	float &elementAt(int row, int col) {return m_mtx[col*4+row];}
	const float &elementAt(int row, int col) const {return m_mtx[col*4+row];}

	float m_mtx[16];
};

struct Point
{
	float x, y, z, w;

	Point();
	Point(const Point &rhs);
	Point(float x, float y, float z);
	~Point();

	void set(float x, float y, float z);

	void multiply(const Point &p, const Matrix &m);

	Point &operator=(const Point &rhs);
	bool operator==(const Point &rhs) const;
};

struct Vector
{
	float x, y, z, w;

	Vector();
	Vector(const Vector &rhs);
	Vector(float x, float y, float z);
	Vector(const Point &from, const Point &to);
	~Vector();

	void set(float x, float y, float z);
	void set(const Point &from, const Point &to);

	void add(const Vector &v);
	void add(const Vector &v1, const Vector &v2);
	void cross(const Vector &v);
	void cross(const Vector &v1, const Vector &v2);
	void divide(const float scalar);
	float dot(const Vector &other) const;
	float dot(const Point &other) const;
	void multiply(const float scalar);
	void multiply(const Vector &v, const Matrix &m);
	void subtract(const Vector &v);
	void subtract(const Vector &v1, const Vector &v2);

	float angle(const Vector &other) const;
	bool isUnitVector() const;
	bool isZeroVector() const;
	float length() const;
	void normalize();
		
	Vector &operator=(const Vector &rhs);
	bool operator==(const Vector &rhs) const;
};

struct Quaternion
{
	float x, y, z, w;

	Quaternion();
	Quaternion(const Quaternion &rhs);
	Quaternion(float x, float y, float z, float w);
	Quaternion(const Vector &axis, float angle);
	Quaternion(float rx, float ry, float rz);
	Quaternion(const Matrix &m);
	~Quaternion();

	void add(const Quaternion &q);
	void add(const Quaternion &q1, const Quaternion &q2);
	void multiply(Quaternion &q);
	void multiply(Quaternion &q1, Quaternion &q2);
	void inverse(const Quaternion &q);
	void conjugate(const Quaternion &q);

	void fromAxisAngle(const Vector &axis, float angle);
	void fromEuler(float rx, float ry, float rz);
	void fromRotationMatrix(const Matrix &m);
	bool isUnitQuaternion() const;
	float length() const;
	void normalize();
	void set(float x, float y, float z, float w);
	void slerp(const Quaternion &from, const Quaternion &to, float t);
	void toAxisAngle(Vector &axis, float &angle) const;
	void toRotationMatrix(Matrix &m) const;

	Quaternion &operator=(const Quaternion &rhs);
	bool operator==(const Quaternion &rhs) const;
};

struct Plane
{
	Vector normal;
	float distOrigin;	// smallest distance from plane to origin

	Plane();
	Plane(const Plane &rhs);
	Plane(const Point &p1, const Point &p2, const Point &p3);
	~Plane();

	void set(const Point &p1, const Point &p2, const Point &p3);
	float distanceTo(const Point &p);

	Plane &operator=(const Plane &rhs);
	bool operator==(const Plane &rhs) const;
};

#endif